@@ -1,19 +1,23 @@
 # -*- coding: utf-8 -*-
 
+
 import re
 import operator
 
+
 def logical_equivalence(*c):
     nonzero_c1 = (c[0] != 0)
     nonzero_c2 = (c[1] != 0)
     return ((nonzero_c1 & nonzero_c2) | (~nonzero_c1 & ~nonzero_c2))
 
+
 # implication
 def logical_implication(*c):
     nonzero_c1 = (c[0] != 0)
     nonzero_c2 = (c[1] != 0)
     return ~(nonzero_c1 & ~nonzero_c2)
 
+
 operators = {'>' : operator.gt,
              '<' : operator.lt,
              '>=': operator.ge,
@@ -23,6 +27,7 @@ operators = {'>' : operator.gt,
              '<->': logical_equivalence,
              '-->': logical_implication}
 
+
 preprocess = {'>':   operator.and_,
               '<':   operator.and_,
               '>=':  operator.and_,
@@ -34,12 +39,14 @@ preprocess = {'>':   operator.and_,
               '<->': operator.or_,
               '-->': operator.or_}
 
+
 logicals = {
         '&': operator.and_,
         '|': operator.or_,
         '^':operator.xor
 }
 
+
 def generate_single_expression(P_columns, Q_columns, pattern, neg = []):
     if pattern == 'percentile':
             expression = '({"' + P_columns[0] + '"} ' + '>=' + ' ' +  str(Q_columns[0])  + ') & ({"' + P_columns[0] + '"} ' + '<=' + ' ' +  str(Q_columns[1])  + ')'
@@ -55,6 +62,7 @@ def generate_single_expression(P_columns, Q_columns, pattern, neg = []):
             else:
                 expression += ' + {"' + P_columns[idx+1]+ '"}'
 
+
         if neg[-1]==-1:
             expression += ' = -{"' + Q_columns[0] + '"})'
         else:
@@ -68,9 +76,11 @@ def generate_single_expression(P_columns, Q_columns, pattern, neg = []):
         expression += " & " + '({"' + P_columns[idx+1]+ '"} ' + pattern + ' {"'+ Q_columns[idx+1] + '"})'
     return expression
 
+
 def generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, parameters):
     '''this function generates the conditional expression from P/Q_columns and P/Q_values'''
 
+
     P_operators = parameters.get("P_operators", ['='] * len(P_columns))
     # operators between Q_column and Q_value, default is =
     Q_operators = parameters.get("Q_operators", ['='] * len(Q_columns))
@@ -79,9 +89,11 @@ def generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, pa
     # logical operators between Q expressions, default is &
     Q_logics = parameters.get("Q_logics", ['&'] * (len(Q_columns) - 1))
 
+
     # Boolean that is set to True if we want to also look for patterns the other way around
     both_ways = parameters.get("both_ways", False)
 
+
     pattern = "IF "
     if isinstance(P_values[0], str):
         pattern += '({"' + str(P_columns[0]) + '"} ' + P_operators[0] + ' "'+ P_values[0] + '")'
@@ -103,6 +115,7 @@ def generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, pa
         else: # numerical value
             pattern += ' ' + Q_logics[idx] + ' ({"' + str(Q_columns[idx+1]) + '"} ' + Q_operators[idx+1] + ' ' + str(Q_values[idx+1]) + ')'
 
+
     if both_ways:
         pattern += " AND IF "
         if isinstance(P_values[0], str):
@@ -125,8 +138,10 @@ def generate_conditional_expression(P_columns, P_values, Q_columns, Q_values, pa
             else: # numerical value
                 pattern += ' ' + Q_logics[idx] + ' ({"' + str(Q_columns[idx+1]) + '"} ' + Q_operators[idx+1] + ' ' + str(Q_values[idx+1]) + ')'
 
+
     return pattern
 
+
 def evaluate_excel_string(s):
     if s != '':
         if type(s)==str:
@@ -136,10 +151,12 @@ def evaluate_excel_string(s):
     else:
         return s
 
+
 def to_xbrl_expressions(pattern, encode, parameters):
     """Placeholder for XBRL"""
     return ["", ""]
 
+
 def replace_and_or(s):
     """Replace and by & and or by |, but not within strings"""
     if re.search(r"(.*?)\'(.*?)\'(.*)", s) is None: # input text does not contain strings
@@ -151,6 +168,7 @@ def replace_and_or(s):
         s = s.replace(item[2], replace_and_or(item[2]))
     return s
 
+
 def replace_div_by_zero(s):
     """Replace division by adding a smal value to numerator"""
     item = re.search(r"({.*?})(/)({.*?})", s)
@@ -161,6 +179,7 @@ def replace_div_by_zero(s):
 def preprocess_pattern(pattern, parameters):
     solvency = parameters.get("solvency", False)
 
+
     pattern = pattern.replace("=" , "==")
     pattern = pattern.replace(">==" , ">=")
     pattern = pattern.replace("<==" , "<=")
@@ -174,6 +193,7 @@ def preprocess_pattern(pattern, parameters):
         pattern = replace_div_by_zero(pattern)
     return pattern
 
+
 def datapoints2pandas(s, encode):
     """Transform datapoints to Pandas datapoints
     Examples:
@@ -190,10 +210,12 @@ def datapoints2pandas(s, encode):
         nonzero_col.append("df["+item+"]")
     return res, nonzero_col
 
+
 def add_brackets(s):
     """Add brackets around expressions with & and |
     """
 
+
     item = re.search(r'(.*)([&|\|])(\s*[(df|df].*)', s)
     if item is not None:
         return '('+add_brackets(item.group(1))+') '+item.group(2).strip()+' ('+add_brackets(item.group(3))+')'
@@ -204,17 +226,23 @@ def add_brackets(s):
         else:
             return s.strip()
 
+
 def expression2pandas(g, nonzero_col, parameters):
     """Transform conditional expression to Pandas code"""
 
+    solvency = parameters.get("solvency", False)
+
     exclude_zero_columns = parameters.get("nonzero", False)
+    notNaN = parameters.get("notNaN", False)
     both_ways = parameters.get("both_ways", False)
+    expres =  parameters.get("expres", False)
     decimal = parameters.get("decimal", 0)
     if re.search('AND IF', g):
         both_ways = True
     if decimal != 0:
         decimal = -decimal
 
+
     item = re.search(r'IF(.*)THEN(.*)', g)
     if item is not None:
         if both_ways:
@@ -223,12 +251,24 @@ def expression2pandas(g, nonzero_col, parameters):
             co_str = 'df[(('+add_brackets(item.group(1))+') & ('+add_brackets(item.group(2))+")) | (~("+add_brackets(item.group(1)) +")& ~("+add_brackets(item.group(2))+ "))]"
             ex_str = 'df[('+add_brackets(item.group(1))+') & ~('+add_brackets(item.group(2))+") | (~("+add_brackets(item.group(1)) +")& ("+add_brackets(item.group(2))+ "))]"
         else:
-            co_str = 'df[('+add_brackets(item.group(1))+') & ('+add_brackets(item.group(2))+")]"
-            ex_str = 'df[('+add_brackets(item.group(1))+') & ~('+add_brackets(item.group(2))+")]"
+            decimal = parameters.get("decimal", 8)
+            if decimal != 0:
+                decimal = -decimal
+            item_cond = re.search(r'(.*)(==)(.*)', item.group(2))
+
+            # only do this when df = df after THEN
+            if item_cond is None or 'df[' not in item_cond.group(3): # take out strings except when string is from sum
+                co_str = 'df[('+add_brackets(item.group(1))+') & ('+add_brackets(item.group(2))+")]"
+                ex_str = 'df[('+add_brackets(item.group(1))+') & ~('+add_brackets(item.group(2))+")]"
+            else:
+                co_str = 'df[('+add_brackets(item.group(1))+') & ('+'abs('+ item_cond.group(1).strip() + '-(' + item_cond.group(3).strip() + '))<1.5e' + str(decimal)+")]"
+                ex_str = 'df[('+add_brackets(item.group(1))+') & ~('+'abs('+ item_cond.group(1).strip() + '-(' + item_cond.group(3).strip() + '))<1.5e' + str(decimal)+")]"
+
     else:
         item = re.search(r'(.*)(==)(.*)', g)
-        if item is None or (re.search(r"'(.*)'", item.group(3)) is not None and 'df[' not in item.group(3)): # take out strings except when string is from sum
 
+        # use decimales only when we have columns or integer after the = sign (and not an expression)
+        if item is None or (re.search(r"'(.*)'", item.group(3)) is not None and 'df[' not in item.group(3)) or expres:
             co_str = 'df[('+add_brackets(g)+')&'
             ex_str = 'df[~('+add_brackets(g)+')&'
             if exclude_zero_columns:
@@ -237,8 +277,8 @@ def expression2pandas(g, nonzero_col, parameters):
                     ex_str += '(' + i +'!=0)&'
             co_str = co_str[:-1] + ']'
             ex_str = ex_str[:-1] + ']'
+
         else:
-            # decimal = parameters.get("decimal", 0)
             g_co = 'abs('+ item.group(1).strip() + '-(' + item.group(3).strip() + '))<1.5e' + str(decimal)
             g_ex = 'abs(' + item.group(1).strip() + '-(' + item.group(3).strip() + '))>=1.5e' + str(decimal)
             co_str = 'df[('+g_co+')&'
@@ -249,4 +289,14 @@ def expression2pandas(g, nonzero_col, parameters):
                     ex_str += '(' + i +'!=0)&'
             co_str = co_str[:-1] + ']'
             ex_str = ex_str[:-1] + ']'
+
+    if notNaN: # If we want to exclude nan values
+        co_str = co_str[:-1]
+        ex_str = ex_str[:-1]
+        all_columns = list(set(re.findall('\["(.*?)"\]', co_str)))
+        for column in all_columns:
+            co_str = co_str + '&(df["' + column +'"].notna())'
+            ex_str = ex_str + '&(df["' + column +'"].notna())'
+        co_str = co_str + ']'
+        ex_str = ex_str + ']'
     return co_str, ex_str
